# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
import codecs
import logging
import sys
from copy import copy
from .grammar import EMPTY, EOF, STOP
from .tables import LALR, SLR, SHIFT, REDUCE, ACCEPT
from .exceptions import ParseError, ParserInitError, DisambiguationError, \
    DynamicDisambiguationConflict, SRConflicts, RRConflicts, \
    expected_symbols_str
from .common import Location, position_context
from .actions import pass_none
from .termui import prints, h_print, a_print
from parglare import termui

if sys.version < '3':
    text = unicode  # NOQA
else:
    text = str


logger = logging.getLogger(__name__)


class Parser(object):
    """Parser works like a DFA driven by LR tables. For a given grammar LR table
    will be created and cached or loaded from cache if cache is found.
    """
    def __init__(self, grammar, in_layout=False, actions=None,
                 layout_actions=None, debug=False, debug_trace=False,
                 debug_colors=False, debug_layout=False, ws='\n\r\t ',
                 build_tree=False, call_actions_during_tree_build=False,
                 tables=LALR, return_position=False,
                 prefer_shifts=None, prefer_shifts_over_empty=None,
                 error_recovery=False, dynamic_filter=None,
                 custom_token_recognition=None, lexical_disambiguation=True,
                 force_load_table=False, table=None):
        self.grammar = grammar
        self.in_layout = in_layout

        EMPTY.action = pass_none
        EOF.action = pass_none
        if actions:
            self.grammar._resolve_actions(action_overrides=actions,
                                          fail_on_no_resolve=True)

        self.layout_parser = None
        if self.in_layout:
            start_production = grammar.get_production_id('LAYOUT')
        else:
            start_production = 1
            layout_symbol = grammar.get_symbol('LAYOUT')
            if layout_symbol:
                self.layout_parser = Parser(
                    grammar,
                    in_layout=True,
                    actions=layout_actions,
                    ws=None, return_position=True,
                    prefer_shifts=True,
                    prefer_shifts_over_empty=True,
                    debug=debug_layout)

        self.ws = ws
        self.return_position = return_position
        self.debug = debug
        self.debug_trace = debug_trace
        self.debug_colors = debug_colors
        termui.colors = debug_colors
        self.debug_layout = debug_layout

        self.build_tree = build_tree
        self.call_actions_during_tree_build = call_actions_during_tree_build

        self.error_recovery = error_recovery
        self.dynamic_filter = dynamic_filter
        self.custom_token_recognition = custom_token_recognition
        self.lexical_disambiguation = lexical_disambiguation

        if table is None:
            from .closure import LR_0, LR_1
            from .tables import create_load_table

            if tables == SLR:
                itemset_type = LR_0
            else:
                itemset_type = LR_1

            if prefer_shifts is None:
                prefer_shifts = True
            if prefer_shifts_over_empty is None:
                prefer_shifts_over_empty = True

            self.table = create_load_table(
                grammar, itemset_type=itemset_type,
                start_production=start_production,
                prefer_shifts=prefer_shifts,
                prefer_shifts_over_empty=prefer_shifts_over_empty,
                lexical_disambiguation=lexical_disambiguation,
                force_load=force_load_table,
                in_layout=self.in_layout)
        else:
            self.table = table

            # warn about overriden parameters
            for name, value, default in [
                ('tables', tables, LALR),
                ('prefer_shifts', prefer_shifts, None),
                ('prefer_shifts_over_empty', prefer_shifts_over_empty, None),
                ('force_load_table', force_load_table, False),
            ]:
                if value is not default:
                    logger.warn("Precomputed table overrides value of "
                                "parameter %s", name)

        self._check_parser()
        if debug:
            self.print_debug()

    def _check_parser(self):
        if self.table.sr_conflicts:
            self.print_debug()
            if self.dynamic_filter:
                unhandled_conflicts = []
                for src in self.table.sr_conflicts:
                    if not src.dynamic:
                        unhandled_conflicts.append(src)
            else:
                unhandled_conflicts = self.table.sr_conflicts

            if unhandled_conflicts:
                raise SRConflicts(unhandled_conflicts)

        # Reduce/Reduce conflicts are fatal for LR parsing
        if self.table.rr_conflicts:
            self.print_debug()
            if self.dynamic_filter:
                unhandled_conflicts = []
                for rrc in self.table.rr_conflicts:
                    if not rrc.dynamic:
                        unhandled_conflicts.append(rrc)
            else:
                unhandled_conflicts = self.table.rr_conflicts

            if unhandled_conflicts:
                raise RRConflicts(unhandled_conflicts)

    def print_debug(self):
        if self.in_layout and self.debug_layout:
            a_print('*** LAYOUT parser ***', new_line=True)
        self.table.print_debug()

    def parse_file(self, file_name, **kwargs):
        """
        Parses content from the given file.
        Args:
            file_name(str): A file name.
        """
        with codecs.open(file_name, 'r', 'utf-8') as f:
            content = f.read()
        return self.parse(content, file_name=file_name, **kwargs)

    def parse(self, input_str, position=0, file_name=None, context=None):
        """
        Parses the given input string.
        Args:
            input_str(str): A string to parse.
            position(int): Position to start from.
            file_name(str): File name if applicable. Used in error reporting.
            context(Context): An object used to keep parser context info.
        """

        if self.debug:
            a_print("*** PARSING STARTED", new_line=True)

        self.errors = []

        next_token = self._next_token
        debug = self.debug
        self.file_name = file_name
        self.in_error_recovery = False

        context = self._get_init_context(context, input_str, position,
                                         file_name)
        assert isinstance(context, Context)

        self._init_dynamic_disambiguation(context)
        self.state_stack = state_stack = [StackNode(context, None)]

        while True:
            cur_state = state_stack[-1].context.state
            if debug:
                a_print("Current state:", str(cur_state.state_id),
                        new_line=True)

            if context.token_ahead is None:
                if not self.in_layout:
                    self._skipws(context)
                    if self.debug:
                        h_print("Layout content:",
                                "'{}'".format(context.layout_content),
                                level=1)

                context.token_ahead = next_token(context)

            if debug:
                h_print("Context:", position_context(context), level=1)
                h_print("Tokens expected:",
                        expected_symbols_str(cur_state.actions.keys()),
                        level=1)
                h_print("Token ahead:", context.token_ahead, level=1)

            actions = cur_state.actions.get(context.token_ahead.symbol)

            if not actions:

                symbols_expected = list(cur_state.actions.keys())
                tokens_ahead = self._get_all_possible_tokens_ahead(context)
                if not self.in_error_recovery:
                    error = self._create_error(
                        context, symbols_expected,
                        tokens_ahead,
                        symbols_before=[cur_state.symbol])
                else:
                    error = self.errors[-1]

                if self.error_recovery:
                    if self._do_recovery(context, error):
                        self.in_error_recovery = True
                        continue

                raise error

            # Dynamic disambiguation
            if self.dynamic_filter:
                actions = self._dynamic_disambiguation(context, actions)

                # If after dynamic disambiguation we still have at least one
                # shift and non-empty reduction or multiple non-empty
                # reductions raise exception.
                if len([a for a in actions
                        if (a.action is SHIFT)
                        or ((a.action is REDUCE) and len(a.prod.rhs))]) > 1:
                    raise DynamicDisambiguationConflict(context, actions)

            # If dynamic disambiguation is disabled either globaly by not
            # giving disambiguation function or localy by not marking
            # any production dynamic for this state take the first action.
            # First action is either SHIFT while there might be empty
            # reductions, or it is the only reduction.
            # Otherwise, parser construction should raise an error.
            act = actions[0]

            if act.action is SHIFT:
                cur_state = act.state

                if debug:
                    a_print("Shift:",
                            "{} \"{}\""
                            .format(cur_state.state_id,
                                    context.token_ahead.value)
                            + " at position " +
                            str(pos_to_line_col(context.input_str,
                                                context.position)), level=1)

                new_position = context.position \
                    + len(context.token_ahead.value)
                context = Context(
                    state=act.state,
                    start_position=context.position,
                    end_position=new_position,
                    token=context.token_ahead,
                    layout_content=context.layout_content_ahead,
                    position=new_position,
                    context=context)

                result = self._call_shift_action(context)
                state_stack.append(StackNode(context, result))

                self.in_error_recovery = False

            elif act.action is REDUCE:
                # if this is EMPTY reduction try to take another if
                # exists.
                if len(act.prod.rhs) == 0:
                    if len(actions) > 1:
                        act = actions[1]
                context.production = production = act.prod

                if debug:
                    a_print("Reducing", "by prod '{}'.".format(production),
                            level=1)

                r_length = len(production.rhs)
                top_stack_context = state_stack[-1].context
                if r_length:
                    start_reduction_context = state_stack[-r_length].context
                    subresults = [x.result for x in state_stack[-r_length:]]
                    del state_stack[-r_length:]
                    cur_state = state_stack[-1].context.state.gotos[
                        production.symbol]
                    context = Context(
                        state=cur_state,
                        start_position=start_reduction_context.start_position,
                        end_position=top_stack_context.end_position,
                        position=top_stack_context.position,
                        production=production,
                        token_ahead=top_stack_context.token_ahead,
                        layout_content=start_reduction_context.layout_content,
                        layout_content_ahead=top_stack_context.layout_content_ahead,  # noqa
                        context=context)
                else:
                    subresults = []
                    cur_state = cur_state.gotos[production.symbol]
                    context = Context(
                        state=cur_state,
                        start_position=context.end_position,
                        end_position=context.end_position,
                        position=context.position,
                        production=production,
                        token_ahead=top_stack_context.token_ahead,
                        layout_content='',
                        layout_content_ahead=top_stack_context.layout_content_ahead,  # noqa
                        context=context)

                # Calling reduce action
                result = self._call_reduce_action(context, subresults)
                state_stack.append(StackNode(context, result))

            elif act.action is ACCEPT:
                if debug:
                    a_print("SUCCESS!!!")
                assert len(state_stack) == 2
                if self.return_position:
                    return state_stack[1].result, context.position
                else:
                    return state_stack[1].result

    def call_actions(self, node, context=None):
        """
        Calls semantic actions for the given tree node.
        """
        self.context = context = context if context else Context()
        context.parser = self

        def set_context(context, node):
            context.start_position = node.start_position
            context.end_position = node.end_position
            context.node = node
            context.production = None
            context.token = None
            context.layout_content = node.layout_content

        def inner_call_actions(node):
            sem_action = node.symbol.action
            if isinstance(node, NodeTerm):
                if sem_action:
                    set_context(context, node)
                    result = sem_action(context, node.value)
                else:
                    result = node.value
            else:
                subresults = []
                # Recursive right to left, bottom up. Simulate LR
                # reductions.
                for n in reversed(node):
                    subresults.append(inner_call_actions(n))
                subresults.reverse()

                if sem_action:
                    set_context(context, node)
                    context.production = node.production
                    assignments = node.production.assignments
                    if assignments:
                        assgn_results = {}
                        for a in assignments.values():
                            if a.op == '=':
                                assgn_results[a.name] = subresults[a.index]
                            else:
                                assgn_results[a.name] = \
                                    bool(subresults[a.index])
                    if type(sem_action) is list:
                        if assignments:
                            result = \
                                sem_action[
                                    node.production.prod_symbol_id](
                                        context, subresults, **assgn_results)
                        else:
                            result = \
                                sem_action[
                                    node.production.prod_symbol_id](context,
                                                                    subresults)
                    else:
                        if assignments:
                            result = sem_action(context, subresults,
                                                **assgn_results)
                        else:
                            result = sem_action(context, subresults)
                else:
                    if len(subresults) == 1:
                        # Unpack if single subresult
                        result = subresults[0]
                    else:
                        result = subresults

            return result

        return inner_call_actions(node)

    def _get_init_context(self, context, input_str, position, file_name):

        context = Context() if not context else context

        context.state = self.table.states[0]
        context.input_str = input_str
        if not hasattr(context, 'file_name') or context.file_name is None:
            context.file_name = file_name
        context.parser = self
        context.position = context.start_position = \
            context.end_position = position
        context.layout_content = ''

        return context

    def _skipws(self, context):

        in_len = len(context.input_str)
        context.layout_content_ahead = ''

        if self.layout_parser:
            _, pos = self.layout_parser.parse(
                context.input_str, context.position, context=copy(context))
            if pos > context.position:
                context.layout_content_ahead = \
                    context.input_str[context.position:pos]
                context.position = pos
        elif self.ws:
            old_pos = context.position
            try:
                while context.position < in_len \
                      and context.input_str[context.position] in self.ws:
                    context.position += 1
            except TypeError:
                raise ParserInitError(
                    "For parsing non-textual content please "
                    "set `ws` to `None`.")
            context.layout_content_ahead = \
                context.input_str[old_pos:context.position]

        if self.debug:
            content = context.layout_content_ahead
            if type(context.layout_content_ahead) is text:
                content = content.replace("\n", "\\n")
            h_print("Skipping whitespaces:",
                    "'{}'".format(content), level=1)
            h_print("New position:", pos_to_line_col(context.input_str,
                                                     context.position),
                    level=1)

    def _next_token(self, context):
        tokens = self._next_tokens(context)
        if not tokens:
            # We have to return something.
            # Returning EMPTY token is not a lie (EMPTY can always be matched)
            # and will cause proper parse error to be raised.
            return EMPTY_token
        elif len(tokens) == 1:
            return tokens[0]
        else:
            raise DisambiguationError(Location(context), tokens)

    def _next_tokens(self, context):
        """
        For the current position in the input stream and actions in the current
        state find next tokens. This function must return only tokens that
        are relevant to specified context - ie it mustn't return a token
        if it's not expected by any action in given state.
        """
        state = context.state
        input_str = context.input_str
        position = context.position
        actions = state.actions
        in_len = len(input_str)
        tokens = []

        # add special tokens (EMPTY, STOP and EOF) if they are applicable
        if EMPTY in actions:
            tokens.append(EMPTY_token)
        if STOP in actions:
            tokens.append(STOP_token)
        if position == in_len:
            tokens.append(EOF_token)
        else:

            # Get tokens by trying recognizers - but only if we are not at
            # the end, because token cannot be empty
            if self.custom_token_recognition:
                def get_tokens():
                    return self._token_recognition(context)

                custom_tokens = self.custom_token_recognition(
                    context, get_tokens,
                )
                if custom_tokens is not None:
                    tokens.extend(custom_tokens)
            else:
                tokens.extend(self._token_recognition(context))

        # do lexical disambiguation if it is enabled
        if self.lexical_disambiguation:
            tokens = self._lexical_disambiguation(context, tokens)

        return tokens

    def _token_recognition(self, context):
        input_str = context.input_str
        actions = context.state.actions
        position = context.position
        finish_flags = context.state.finish_flags

        tokens = []
        last_prior = -1
        for idx, symbol in enumerate(actions):
            if symbol.prior < last_prior and tokens:
                break
            last_prior = symbol.prior
            try:
                tok = symbol.recognizer(input_str, position)
            except TypeError:
                tok = symbol.recognizer(context, input_str, position)
            if tok:
                tokens.append(Token(symbol, tok))
                if finish_flags[idx]:
                    break
        return tokens

    def _get_all_possible_tokens_ahead(self, context):
        """
        Check what is ahead no matter the current state.
        Just check with all recognizers available.
        """
        tokens = []
        if context.position < len(context.input_str):
            for terminal in self.grammar.terminals.values():
                try:
                    tok = terminal.recognizer(context.input_str,
                                              context.position)
                except TypeError:
                    tok = terminal.recognizer(context, context.input_str,
                                              context.position)
                if tok:
                    tokens.append(Token(terminal, tok))
        return tokens

    def _init_dynamic_disambiguation(self, context):
        if self.dynamic_filter:
            if self.debug:
                prints("\tInitializing dynamic disambiguation.")
            self.dynamic_filter(context, None, None)

    def _dynamic_disambiguation(self, context, actions):

        dyn_actions = []
        for a in actions:
            if a.action is SHIFT:
                if self._call_dynamic_filter(context, SHIFT, None):
                    dyn_actions.append(a)
            elif a.action is REDUCE:
                r_len = len(a.prod.rhs)
                if r_len:
                    subresults = [x.result
                                  for x in self.state_stack[-r_len:]]
                else:
                    subresults = []
                context.production = a.prod
                if self._call_dynamic_filter(context, REDUCE, subresults):
                    dyn_actions.append(a)
            else:
                dyn_actions.append(a)
        return dyn_actions

    def _call_dynamic_filter(self, context, action, subresults):
        if (action is SHIFT and not context.token_ahead.symbol.dynamic)\
           or (action is REDUCE and not context.production.dynamic):
            return True

        if self.debug:
            h_print("Calling filter for action:",
                    " {}, token={}{}{}"
                    .format(
                        "SHIFT" if action is SHIFT else "REDUCE",
                        context.token,
                        ", prod={}".format(context.production)
                        if action is REDUCE else "",
                        ", subresults={}".format(subresults)
                        if action is REDUCE else ""), level=2)

        accepted = self.dynamic_filter(context, action, subresults)
        if self.debug:
            if accepted:
                a_print("Action accepted.", level=2)
            else:
                a_print("Action rejected.", level=2)

        return accepted

    def _call_shift_action(self, context):
        """
        Calls registered shift action for the given grammar symbol.
        """
        debug = self.debug
        token = context.token
        sem_action = token.symbol.action

        if self.build_tree:
            # call action for building tree node if tree building is enabled
            if debug:
                h_print("Building terminal node",
                        "'{}'.".format(token.symbol.name), level=2)

            # If both build_tree and call_actions_during_build are set to
            # True, semantic actions will be call but their result will be
            # discarded. For more info check following issue:
            # https://github.com/igordejanovic/parglare/issues/44
            if self.call_actions_during_tree_build and sem_action:
                sem_action(context, token.value)

            return treebuild_shift_action(context)

        if sem_action:
            result = sem_action(context, token.value)

        else:
            if debug:
                h_print("No action defined",
                        "for '{}'. "
                        "Result is matched string.".format(token.symbol.name),
                        level=1)
            result = token.value

        if debug:
            h_print("Action result = ",
                    "type:{} value:{}"
                    .format(type(result), repr(result)), level=1)

        return result

    def _call_reduce_action(self, context, subresults):
        """
        Calls registered reduce action for the given grammar symbol.
        """
        debug = self.debug
        result = None
        bt_result = None
        production = context.production

        if self.build_tree:
            # call action for building tree node if enabled.
            if debug:
                h_print("Building non-terminal node",
                        "'{}'.".format(production.symbol.name), level=2)

            bt_result = treebuild_reduce_action(context, nodes=subresults)
            if not self.call_actions_during_tree_build:
                return bt_result

        sem_action = production.symbol.action
        if sem_action:
            assignments = production.assignments
            if assignments:
                assgn_results = {}
                for a in assignments.values():
                    if a.op == '=':
                        assgn_results[a.name] = subresults[a.index]
                    else:
                        assgn_results[a.name] = bool(subresults[a.index])

            if type(sem_action) is list:
                if assignments:
                    result = sem_action[production.prod_symbol_id](
                        context, subresults, **assgn_results)
                else:
                    result = sem_action[production.prod_symbol_id](context,
                                                                   subresults)
            else:
                if assignments:
                    result = sem_action(context, subresults, **assgn_results)
                else:
                    result = sem_action(context, subresults)

        else:
            if debug:
                h_print("No action defined",
                        " for '{}'.".format(production.symbol.name), level=1)
            if len(subresults) == 1:
                if debug:
                    h_print("Unpacking a single subresult.", level=1)
                result = subresults[0]
            else:
                if debug:
                    h_print("Result is a list of subresults.", level=1)
                result = subresults

        if debug:
            h_print("Action result =",
                    "type:{} value:{}"
                    .format(type(result), repr(result)), level=1)

        # If build_tree is set to True, discard the result of the semantic
        # action, and return the result of treebuild_reduce_action.
        return bt_result if bt_result is not None else result

    def _lexical_disambiguation(self, context, tokens):
        """
        For the given list of matched tokens apply disambiguation strategy.

        Args:
        tokens (list of Token)
        """

        if self.debug:
            h_print("Lexical disambiguation.",
                    " Tokens: {}".format([x for x in tokens]), level=1)

        if len(tokens) <= 1:
            return tokens

        # prefer STOP over EMPTY and EOF
        if STOP_token in tokens:
            tokens = [t for t in tokens if t not in (EMPTY_token, EOF_token)]
        # prefer EMTPY over EOF
        elif EMPTY_token in tokens:
            tokens = [t for t in tokens if t != EOF_token]

        # Longest-match strategy.
        max_len = max((len(x.value) for x in tokens))
        tokens = [x for x in tokens if len(x.value) == max_len]
        if self.debug:
            h_print("Disambiguation by longest-match strategy.",
                    "Tokens: {}".format([x for x in tokens]), level=1)
        if len(tokens) == 1:
            return tokens

        # try to find preferred token.
        pref_tokens = [x for x in tokens if x.symbol.prefer]
        if pref_tokens:
            if self.debug:
                h_print("Preferring tokens {}.".format(pref_tokens),
                        level=1)
            return pref_tokens

        return tokens

    def _do_recovery(self, context, error):

        debug = self.debug
        if debug:
            a_print("**Recovery initiated.**")

        if type(self.error_recovery) is bool:
            # Default recovery
            if debug:
                prints("\tDoing default error recovery.")
            token, position = self.default_error_recovery(context)
        else:
            # Custom recovery provided during parser construction
            if debug:
                prints("\tDoing custom error recovery.")
            token, position = self.error_recovery(context, error)

        # The recovery may either decide to skip erroneous part of
        # the input and resume at the place that can continue or it
        # might decide to fill in missing tokens.
        if position:
            last_error = self.errors[-1]
            last_error.location.end_position = position
            context.position = position
            if debug:
                h_print("Advancing position to ",
                        pos_to_line_col(context.input_str, position),
                        level=1)

        context.token_ahead = token
        if token and debug:
            h_print("Introducing token {}", repr(token), level=1)

        return bool(token or position)

    def default_error_recovery(self, context):
        """The default recovery strategy is to drop char/object at current position
        and try to continue.

        Args:
            context(Context): The parsing context

        Returns:
            (None for new Token, new position)

        """
        return None, context.position + 1 \
            if context.position < len(context.input_str) else None

    def _create_error(self, context, symbols_expected, tokens_ahead,
                      symbols_before):
        context = copy(context)
        context.start_position = context.position
        context.end_position = context.position
        error = ParseError(Location(context=context),
                           symbols_expected,
                           tokens_ahead,
                           symbols_before=symbols_before)

        if self.debug:
            a_print("Error: ", error, level=1)

        self.errors.append(error)

        return error


class Context:
    """
    Args:
        state(LRState):
        file_name(str):
        position(int):
        start_position, end_position(int):
        layout_content(str): Layout content preceeding current token.
        layout_content_ahead(str): Layout content preceeding token_ahead.
        token(Token): Token for shift operation.
        token_ahead(Token): Token recognized ahead at position in given
             state.
        extra(anything): Used for additional state maintained by the user.
             If not given empty dict is used.
    """

    __local = ['state',
               'position',
               'start_position',
               'end_position',
               'token',
               'token_ahead',
               'production',
               'layout_content',
               'layout_content_ahead',
               'node']
    __t = ['file_name',
           'input_str',
           'parser',
           'extra']

    __slots__ = __local + __t

    def __init__(self, state=None, position=None, start_position=None,
                 end_position=None, token=None, token_ahead=None,
                 production=None, layout_content=None,
                 layout_content_ahead=None, node=None, file_name=None,
                 input_str=None, parser=None, extra=None, context=None):
        self.state = state
        self.position = position
        self.start_position = start_position
        self.end_position = end_position
        self.token = token
        self.token_ahead = token_ahead
        self.production = production
        self.layout_content = layout_content
        self.layout_content_ahead = layout_content_ahead
        self.node = node
        if context:
            self.extra = context.extra
            self.file_name = context.file_name
            self.input_str = context.input_str
            self.parser = context.parser
        else:
            self.extra = extra if extra is not None else {}
            self.file_name = file_name
            self.input_str = input_str
            self.parser = parser

    @property
    def symbol(self):
        if self.token is not None:
            return self.token.symbol
        elif self.production is not None:
            return self.production.symbol
        elif self.node is not None:
            return self.node.symbol

    def __str__(self):
        if self.symbol:
            return str(self.symbol)
        elif self.token:
            return str(self.token)
        else:
            return str(self.production)


class StackNode:
    __slots__ = ['context', 'result']

    def __init__(self, context, result):
        self.context = context
        self.result = result

    def __repr__(self):
        return "<StackNode({}, pos=({}-{}))>"\
            .format(repr(self.context.state),
                    self.context.start_position, self.context.end_position)


class Node(object):
    """A node of the parse tree."""

    __slots__ = ['start_position', 'end_position', 'layout_content']

    def __init__(self, start_position, end_position, layout_content=None):
        self.start_position = start_position
        self.end_position = end_position
        self.layout_content = layout_content

    def __repr__(self):
        return str(self)

    def __iter__(self):
        return iter([])

    def __reversed__(self):
        return iter([])


class NodeNonTerm(Node):
    __slots__ = ['production', 'children']

    def __init__(self, start_position, end_position, production, children,
                 layout_content=None):
        super(NodeNonTerm, self).__init__(start_position,
                                          end_position,
                                          layout_content=layout_content)
        self.production = production
        self.children = children

    def tree_str(self, depth=0):
        indent = '  ' * depth
        s = '{}[{}->{}]'.format(self.production.symbol,
                                self.start_position,
                                self.end_position)
        if self.children:
            for n in self.children:
                if hasattr(n, 'tree_str'):
                    s += '\n' + indent + n.tree_str(depth+1)
                else:
                    s += '\n' + indent + n.__class__.__name__ \
                         + '(' + str(n) + ')'
        return s

    @property
    def symbol(self):
        return self.production.symbol

    def __str__(self):
        return '<NonTerm(start={}, end={}, sym={})>'\
            .format(self.start_position, self.end_position,
                    self.production.symbol)

    def __iter__(self):
        return iter(self.children)

    def __reversed__(self):
        return reversed(self.children)


class NodeTerm(Node):
    __slots__ = ['token']

    def __init__(self, start_position, end_position, token,
                 layout_content=None):
        super(NodeTerm, self).__init__(start_position,
                                       end_position,
                                       layout_content=layout_content)
        self.token = token

    @property
    def symbol(self):
        return self.token.symbol

    @property
    def value(self):
        return self.token.value

    def tree_str(self, depth=0):
        return '{}[{}->{}, "{}"]'.format(self.symbol,
                                         self.start_position,
                                         self.end_position,
                                         self.value)

    def __str__(self):
        return '<Term(start={}, end={}, sym={}, val="{}")>'\
            .format(self.start_position, self.end_position, self.symbol,
                    self.value[:20])

    def __iter__(self):
        return iter([])

    def __reversed__(self):
        return iter([])


class Token(object):
    """
    Token or lexeme matched from the input.
    """
    __slots__ = ['symbol', 'value', 'length']

    def __init__(self, symbol=None, value='', length=None):
        self.symbol = symbol
        self.value = value
        self.length = length if length is not None else len(value)

    def __repr__(self):
        return "<{}({})>".format(text(self.symbol), text(self.value))

    def __len__(self):
        return self.length

    def __bool__(self):
        return True


STOP_token = Token(STOP)
EMPTY_token = Token(EMPTY)
EOF_token = Token(EOF)


def treebuild_shift_action(context):
    return NodeTerm(context.start_position, context.end_position,
                    context.token, context.layout_content)


def treebuild_reduce_action(context, nodes):
    if nodes:
        return NodeNonTerm(nodes[0].start_position, nodes[-1].end_position,
                           context.production, nodes, context.layout_content)
    else:
        return NodeNonTerm(context.start_position, context.end_position,
                           context.production, nodes, context.layout_content)


def pos_to_line_col(input_str, position):
    """
    Returns position in the (line,column) form.
    """

    if position is None:
        return None, None

    if type(input_str) is not text:
        # If we are not parsing string
        return 1, position

    line = 1
    old_pos = 0
    try:
        cur_pos = input_str.index("\n")
        while cur_pos < position:
            line += 1
            old_pos = cur_pos + 1
            cur_pos = input_str.index("\n", cur_pos + 1)
    except ValueError:
        pass

    return line, position - old_pos



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import os
from parglare import Grammar
from parglare import Parser


def test_grammar_with_unicode():
    this_folder = os.path.dirname(__file__)
    grammar = Grammar.from_file(os.path.join(this_folder, "names.pg"))
    parser = Parser(grammar)
    inp = 'МИША МЫЛ РАМУ'
    result = parser.parse(inp)
    assert result



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
import codecs
from itertools import chain, takewhile
from copy import copy
from parglare import Parser
from parglare import termui as t
from .exceptions import ParseError
from .parser import SHIFT, REDUCE, ACCEPT, pos_to_line_col, Context, Token
from .common import Location, position_context
from .common import replace_newlines as _
from .tables import LALR
from .export import dot_escape
from .termui import prints, h_print, a_print


def no_colors(f):
    """
    Decorator for trace methods to prevent ANSI COLOR codes apearing in
    the trace dot output.
    """
    def nc_f(*args, **kwargs):
        self = args[0]
        t.colors = False
        r = f(*args, **kwargs)
        t.colors = self.debug_colors
        return r
    return nc_f


class GLRParser(Parser):
    """
    A Tomita-style GLR parser.
    """
    def __init__(self, grammar, actions=None,
                 layout_actions=None, debug=False, debug_trace=False,
                 debug_colors=False, debug_layout=False, ws='\n\r\t ',
                 build_tree=False, call_actions_during_tree_build=False,
                 tables=LALR, return_position=False,
                 prefer_shifts=None, prefer_shifts_over_empty=None,
                 error_recovery=False, dynamic_filter=None,
                 custom_token_recognition=None, lexical_disambiguation=None,
                 force_load_table=False, table=None, **kwargs):

        if table is None:
            # The default for GLR is not to use any strategy preferring shifts
            # over reduce thus investigating all possibilitites.
            # These settings are only applicable if parse table is not computed
            # yet. If it is, then leave None values to avoid
            # "parameter overriden" warnings.
            prefer_shifts = False \
                if prefer_shifts is None else prefer_shifts
            prefer_shifts_over_empty = False \
                if prefer_shifts_over_empty is None \
                else prefer_shifts_over_empty
            if lexical_disambiguation is None:
                lexical_disambiguation = False

        super(GLRParser, self).__init__(
            grammar=grammar,
            actions=actions, layout_actions=layout_actions,
            debug=debug, debug_trace=debug_trace,
            debug_colors=debug_colors, debug_layout=debug_layout, ws=ws,
            build_tree=build_tree,
            call_actions_during_tree_build=call_actions_during_tree_build,
            tables=tables, return_position=return_position,
            prefer_shifts=prefer_shifts,
            prefer_shifts_over_empty=prefer_shifts_over_empty,
            error_recovery=error_recovery, dynamic_filter=dynamic_filter,
            custom_token_recognition=custom_token_recognition,
            lexical_disambiguation=lexical_disambiguation,
            force_load_table=force_load_table, table=table, **kwargs)

    def _check_parser(self):
        """
        Conflicts in table are allowed with GLR.
        """
        pass

    def parse(self, input_str, position=0, file_name=None, context=None):
        """
        Parses the given input string.
        Args:
            input_str(str): A string to parse.
            position(int): Position to start from.
            file_name(str): File name if applicable. Used in error reporting.
            context(Context): An object used to keep parser context info.
        """

        if self.debug:
            a_print("*** PARSING STARTED\n")
            self.debug_step = 0
            if self.debug_trace:
                self.dot_trace = ""

        self.errors = []
        self.in_error_recovery = None
        self.in_error_reporting = False
        self.last_position = 0
        self.expected = set()
        self.empty_reductions_results = {}
        self.heads_for_recovery = []
        self.last_heads_for_reduce = []
        self.last_shifts = {}
        self.reducing_heads = []
        self.heads_for_shift = []
        self.finish_head = None

        self.file_name = file_name

        self.context = context = self._get_init_context(context, input_str,
                                                        position, file_name)
        assert isinstance(context, Context)

        self._init_dynamic_disambiguation(context)
        self._skipws(context)

        # We start with a single parser head in state 0.
        start_head = GSSNode(context, number_of_trees=1)
        self.heads_for_reduce = [start_head]

        if self.debug and self.debug_trace:
            self._trace_head(start_head,
                             str(start_head.context.state.state_id))

        # The main loop
        while self.heads_for_reduce:

            self.last_heads_for_reduce = list(self.heads_for_reduce)

            self._do_reductions()
            if self.heads_for_shift:
                self._do_shifts()

            # If after shifting we don't have any heads for reduce and we
            # haven't found any final parse do error reporting.
            if not self.heads_for_reduce and not self.finish_head:
                self.in_error_reporting = True
                try:
                    if self.debug:
                        a_print("*** ENTERING ERROR REPORTING MODE.",
                                new_line=True)

                    self._setup_error_reporting()
                    self._do_reductions()
                finally:
                    self.in_error_reporting = False
                    if self.debug:
                        a_print("*** LEAVING ERROR REPORTING MODE.",
                                new_line=True)
                        h_print("Tokens expected:", self.expected, level=1)
                        h_print("Tokens found:", self.tokens_ahead, level=1)

            # After error reporing do error recovery if enabled.
            if self.error_recovery:
                if not self.heads_for_reduce and not self.finish_head:
                    if self.heads_for_recovery:
                        if not self.in_error_recovery:
                            context = self.heads_for_recovery[0].context
                            error = self._create_error(
                                context, self.expected,
                                self.tokens_ahead,
                                list({h.context.state.symbol
                                      for h in self.last_heads_for_reduce}))
                        else:
                            error = self.errors[-1]
                        if self._do_recovery(error):
                            self.in_error_recovery = True
                            continue
                else:
                    # Get out from error recovery mode.
                    self.in_error_recovery = False
                    # Report dying heads
                    if self.debug:
                        for h in self.heads_for_recovery:
                            a_print("{}. Killing head:"
                                    .format(self.debug_step), h,
                                    level=1)
                            if self.debug_trace:
                                self._trace_step_kill(h)
                                self.debug_step += 1

        if not self.finish_head:
            if self.debug and self.debug_trace:
                self._export_dot_trace()
            context = self.context
            context.start_position = context.end_position = context.position
            last_heads_for_reduce = self.last_heads_for_reduce
            self._remove_transient_state()
            raise ParseError(Location(context=context),
                             self.expected, self.tokens_ahead,
                             list({h.context.state.symbol
                                   for h in last_heads_for_reduce}))

        results = [x[1] for x in self.finish_head.parents]
        self._remove_transient_state()
        if self.debug:
            a_print("*** {} sucessful parse(s).".format(len(results)))
            if self.debug_trace:
                self._export_dot_trace()

        return results

    def _do_reductions(self):
        """
        Reduces active heads until no more heads can be reduced.
        """
        debug = self.debug
        if debug:
            a_print("**REDUCING HEADS", new_line=True)
            self._debug_active_heads(self.heads_for_reduce)

        next_tokens = self._next_tokens
        reduce = self._reduce

        # Reductions
        heads_for_reduce = self.heads_for_reduce
        self.heads_for_shift = []

        # For automata loop detection
        self.reducing_heads = []

        if self.error_recovery and not self.in_error_reporting:
            self.heads_for_recovery = []

        while heads_for_reduce:
            head = heads_for_reduce.pop()
            self.reducing_heads.append(head)
            if debug:
                a_print("Reducing head: ", str(head), new_line=True)

            self.context = context = head.context
            position = context.position
            actions = context.state.actions
            token = context.token_ahead

            if token is None:
                if debug:
                    h_print("Finding lookaheads.", level=1)

                self._skipws(context)
                if position > self.last_position:
                    self.last_position = position
                    self.expected = set()

                if debug:
                    self._debug_context(context, token,
                                        expected_symbols=actions.keys())

                tokens = next_tokens(context)

                if debug:
                    h_print("Token(s) ahead: ", _(str(tokens)), level=1)

                if not tokens:
                    if self.error_recovery:
                        # collect for possible recovery
                        self.heads_for_recovery.append(head)
                    elif debug:
                        a_print("** Killing head: ", head, level=1)
                        if self.debug_trace:
                            self._trace_step_kill(head)
                else:
                    for idx, token in enumerate(tokens):
                        reduce_head = head.for_token(token)
                        self.heads_for_reduce.insert(0, reduce_head)
                    continue

            else:
                symbol_actions = actions.get(token.symbol, [])
                for symbol_action in symbol_actions:
                    action = symbol_action.action

                    if action is ACCEPT:
                        if self.in_error_reporting:
                            self.expected.add(token.symbol)
                        else:
                            if debug:
                                a_print("*** {}. SUCCESS!!!!", self.debug_step)
                                self.debug_step += 1
                                if self.debug_trace:
                                    self._trace_step_finish(head)
                            if self.finish_head:
                                self.finish_head.merge_head(head, self)
                            else:
                                self.finish_head = head

                        # This break is supicious.
                        # It prevents from ivestigating posibilities past
                        # accepting state. Without it test_cyclic_grammar_1
                        # fails returning more than one parse.
                        break

                    if action is REDUCE:
                        reduce(head, symbol_action.prod)

                    if action is SHIFT:
                        if self.in_error_reporting:
                            self.expected.add(token.symbol)
                        else:
                            self._add_to_heads_for_shift(head)

                if debug:
                    h_print("No more reductions for this head and lookahead"
                            " token:", _(str(token)), level=1, new_line=True)

    def _do_shifts(self):
        """Perform all shifts.

        Heads that are shifted successfully will be new candidates for
        reducing. If head is not shifted we have a dying head. They will be
        collected for error recovery if enabled.

        """
        debug = self.debug
        if self.debug:
            a_print("**SHIFTING HEADS", new_line=True)

        heads_for_shift = self.heads_for_shift
        self.last_shifts = {}

        if self.debug:
            self._debug_active_heads(heads_for_shift)

        for head in heads_for_shift:
            if debug:
                a_print("Shifting head: ", head, new_line=True)

            self.context = context = head.context

            if debug:
                self._debug_context(context, context.token_ahead,
                                    expected_symbols=None)

            # First action should be SHIFT if it is possible to shift by this
            # token.
            action = context.state.actions.get(context.token_ahead.symbol,
                                               [None])[0]
            if action and action.action is SHIFT:
                if self.dynamic_filter and \
                   not self._call_dynamic_filter(context, SHIFT, None):
                    pass
                else:
                    self._shift(head, action.state, context)
            else:
                # This should never happen as the shift possibility is checked
                # during reducing and only those heads that can be shifted are
                # appended to heads_for_shift
                assert False, "No shift operation possible."

    def _reduce(self, head, production):
        """Executes reduce operation for the given head and production.
        """
        debug = self.debug
        self.context = context = head.context

        if debug:
            a_print("{}. REDUCING by prod ".format(self.debug_step),
                    production, level=1, new_line=True)
            self.debug_step += 1

        prod_len = len(production.rhs)
        roots = []
        if not prod_len:
            context = Context(
                state=context.state.gotos[production.symbol],
                production=production,
                start_position=context.start_position,
                end_position=context.start_position,
                position=context.position,
                layout_content=context.layout_content,
                token_ahead=context.token_ahead,
                layout_content_ahead=context.layout_content_ahead,
                context=context)

            if self.dynamic_filter and \
               not self._call_dynamic_filter(context, REDUCE, []):
                pass
            else:
                new_head = GSSNode(context)
                self._merge_create_head(new_head, head, head, [], True, True)
        else:
            # Find roots of new heads by going backwards for prod_len steps
            # following all possible paths.
            # Collect subresults along the way to be used with semantic actions
            to_process = [(head, context, [], prod_len, False, True)]
            if debug:
                h_print("Calculate reduction paths of length {}, "
                        "choose only non-empty if possible:"
                        .format(prod_len), level=1)
                h_print("start node=",
                        "[{}], symbol={}, empty=[{},{}], "
                        "length={}".format(head, context.state.symbol,
                                           head.any_empty,
                                           head.all_empty, prod_len), level=2)
            roots = []
            while to_process:
                node, first_head_context, subresults, length, path_has_empty,\
                    path_all_empty = to_process.pop()
                length = length - 1
                if debug:
                    h_print("node = {}".format(node), level=2, new_line=True)
                    h_print("backpath length = {}{}"
                            .format(prod_len - length,
                                    " - ROOT" if not length else ""),
                            level=2)
                for parent, res, any_empty, all_empty in node.parents:
                    path_has_empty = path_has_empty or any_empty
                    path_all_empty = path_all_empty and all_empty
                    if debug:
                        h_print("", "parent=[{}]: path_any_empty={}, "
                                "path_all_empty={}"
                                .format(parent,
                                        path_has_empty, path_all_empty),
                                level=3)
                    parent_subres = [res] + subresults
                    if length:
                        to_process.append(
                            (parent, node.context, parent_subres, length,
                             path_has_empty, path_all_empty))
                    else:
                        roots.append((parent, node.context,
                                      parent_subres, path_has_empty,
                                      path_all_empty))

            # Favour non-empty paths if exists or partialy empty.
            # In none of those exist use empty paths.
            non_empty = [r for r in roots if not r[3] and not r[4]]
            if non_empty:
                roots = non_empty
            else:
                non_empty = [r for r in roots if not r[4]]
                if non_empty:
                    roots = non_empty

            if debug:
                h_print("Reduction paths = ", len(roots),
                        level=1, new_line=True)
                h_print("Roots:", level=1)
                for idx, r in enumerate(roots):
                    h_print("{}.".format(idx+1), r[0], level=2)

            # Create new heads.
            for idx, (root, first_head_context, subresults,
                      any_empty, all_empty) in enumerate(roots):
                if debug:
                    h_print("Reducing path {}:".format(idx + 1),
                            level=1, new_line=True)

                context = Context(
                    state=root.context.state.gotos[production.symbol],
                    production=production,
                    start_position=first_head_context.start_position,
                    end_position=context.end_position,
                    position=context.position,
                    layout_content=first_head_context.layout_content,
                    token_ahead=context.token_ahead,
                    layout_content_ahead=context.layout_content_ahead,
                    context=context)

                if self.dynamic_filter and \
                   not self._call_dynamic_filter(context, REDUCE,
                                                 subresults):
                    pass
                else:
                    new_head = GSSNode(context)
                    self._merge_create_head(new_head, head, root, subresults,
                                            any_empty, all_empty)
                if debug:
                    print()

    def _shift(self, head, state, context):
        """Execute shift operation at the given position to the given state.

        Shift token determined by the given state from input at given position
        and create the new head. Parents will be all nodes that had shift
        action with the given state at the given position.
        """

        last_shifts = self.last_shifts
        debug = self.debug
        token = context.token_ahead

        shifted_head = last_shifts.get((state.state_id,
                                        context.position, token.symbol),
                                       None)
        if shifted_head:
            # If this token has already been shifted connect
            # shifted head to this head.
            result = shifted_head.parents[0][1]
            shifted_head.create_link(head, result, False, False, self)
            if debug and self.debug_trace:
                self._trace_step(head, shifted_head, head,
                                 "S:{}({})".format(
                                     dot_escape(token.symbol.name),
                                     dot_escape(token.value)))
        else:

            if self.debug:
                a_print("{}. SHIFTING".format(self.debug_step),
                        _("\"{}\" to state {} "
                          .format(token.value, context.state.state_id) +
                          "at position " +
                          str(pos_to_line_col(context.input_str,
                                              context.start_position))),
                        level=1, new_line=True)
                self.debug_step += 1

            position = context.position + len(token)
            context = Context(state=state,
                              token=token,
                              start_position=context.position,
                              end_position=position,
                              layout_content=context.layout_content_ahead,
                              position=position,
                              context=context)

            result = self._call_shift_action(context)

            new_head = GSSNode(context)

            # Cache this shift for further shift of the same symbol on the same
            # position.
            last_shifts[(state.state_id, context.start_position,
                         token.symbol)] = new_head

            self.heads_for_reduce.append(new_head)
            if debug:
                a_print("New shifted head ", new_head, level=1)
                if self.debug_trace:
                    self._trace_head(new_head,
                                     "{}:{}".format(
                                         state.state_id,
                                         dot_escape(state.symbol.name)))
                    self._trace_step(head, new_head, head,
                                     "S:{}({})".format(
                                         dot_escape(token.symbol.name),
                                         dot_escape(token.value)))

            new_head.create_link(head, result, False, False, self)

    def _add_to_heads_for_shift(self, new_head):
        """Adds new head for shift or merges if already added."""
        for head in self.heads_for_shift:
            if head == new_head:
                if self.debug:
                    h_print("Merging head for shifting.", level=1)
                head.merge_head(new_head, self)
                break
        else:
            if self.debug:
                h_print("New head for shifting: ", new_head,
                        level=1, new_line=True)
            self.heads_for_shift.append(new_head)

    def _merge_create_head(self, new_head, old_head, root_head, subresults,
                           any_empty, all_empty):
        """Adds new head or merges if already exist on the stack. Executes semantic
        actions. Detects automata looping.
        """

        debug = self.debug
        context = new_head.context

        if new_head == old_head:
            # Special case is reduction of empty production. For automata state
            # self-reference create stack node loop.
            if debug:
                a_print("Looping automata transition.", level=1)
            result = self._call_reduce_action(context, subresults)
            old_head.parents.append((old_head, result, True, True))

        if all_empty and new_head in self.reducing_heads:
            # Detect automata loop. If we are reducing to the head we already
            # had and the new head is empty we have a loop due to EMPTY
            # reductions.
            if debug:
                a_print("Automata loop detected. ",
                        "Rejecting the new head: {}".format(str(new_head)),
                        level=1)
                if self.debug_trace:
                    self._trace_step_kill(old_head)
            return

        result = self._call_reduce_action(context, subresults)

        for head in chain(self.heads_for_reduce,
                          [self.finish_head] if self.finish_head else []):
            if head == new_head:
                new_head.create_link(root_head, result, any_empty, all_empty,
                                     self)
                if head.merge_head(new_head, self):
                    if self.debug and self.debug_trace:
                        self._trace_step(
                            old_head, head, root_head,
                            "R:{}".format(dot_escape(context.production)))
                break
        else:
            self.heads_for_reduce.append(new_head)
            if self.debug:
                a_print("New reduced head ", new_head, level=2, new_line=True)
                if self.debug_trace:
                    self._trace_head(new_head, "{}:{}".format(
                        new_head.context.state.state_id,
                        dot_escape(new_head.context.state.symbol.name)))
            new_head.create_link(root_head, result, any_empty, all_empty, self)

            if self.debug and self.debug_trace:
                self._trace_step(old_head, new_head, root_head,
                                 "R:{}".format(dot_escape(context.production)))

    def _setup_error_reporting(self):
        """
        To correctly report what is found ahead and what is expected we shall:
        - execute all grammar recognizers at the farther position reached in
          the input by the last shifted heads. This will be part of the error
          report (what is found ahead if anything can be recognized).
        - for all last shifted heads, simulate parsing for each of possible
          lookaheads in the head's state until either SHIFT or ACCEPT is
          successfuly executed. Collect each possible lookahead where this is
          achieved for reporting. This will be another part of the error
          report (what is expected).
        """
        # Start with the last shifted heads sorted by position.
        self.last_heads_for_reduce.sort(key=lambda h: h.context.position,
                                        reverse=True)
        context = self.last_heads_for_reduce[0].context
        farthest_heads = takewhile(
            lambda h: h.context.position == context.position,
            self.last_heads_for_reduce)

        self.tokens_ahead = self._get_all_possible_tokens_ahead(context)

        for head in farthest_heads:
            for possible_lookahead in head.context.state.actions.keys():
                self.heads_for_reduce.append(
                    head.for_token(Token(possible_lookahead, [])))

    def _do_recovery(self, error):
        """If recovery is enabled, does error recovery for the heads in
        heads_for_recovery.

        """
        debug = self.debug
        for head in self.heads_for_recovery:
            context = head.context
            input_str = context.input_str
            symbols = context.state.actions.keys()
            if debug:
                a_print("**Error found. ",
                        "Recovery initiated for head {}.".format(head),
                        level=1, new_line=True)
                h_print("Symbols expected: ", symbols, level=1)
            if type(self.error_recovery) is bool:
                # Default recovery
                if debug:
                    prints("\tDoing default error recovery.")
                token, position = self.default_error_recovery(context)
            else:
                # Custom recovery provided during parser construction
                if debug:
                    prints("\tDoing custom error recovery.")
                token, position = self.error_recovery(context, error)

            if position is not None or token is not None:
                if position:
                    last_error = self.errors[-1]
                    last_error.location.end_position = position
                    context.position = position
                    if debug:
                        h_print("Advancing position to ",
                                pos_to_line_col(input_str, position),
                                level=1)
                context.token_ahead = token

                if token and debug:
                    h_print("Introducing token {}", repr(token), level=1)

                self.heads_for_reduce.append(head)

            else:
                if debug:
                    a_print("Killing head: ", head, level=1)
                    if self.debug_trace:
                        self._trace_step_kill(head)

        return bool(token or position)

    def _remove_transient_state(self):
        """
        Delete references to transient parser objects to lower memory
        consumption.
        """
        del self.finish_head
        del self.empty_reductions_results
        del self.heads_for_recovery
        del self.heads_for_reduce
        del self.heads_for_shift
        del self.last_heads_for_reduce
        del self.last_shifts
        del self.reducing_heads

    def _debug_active_heads(self, heads):
        h_print("Active heads = ", len(heads))
        for head in heads:
            prints("\t{}".format(head))
        h_print("Number of trees = ", sum([h.number_of_trees for h in heads]))

    def _debug_context(self, context, lookahead_tokens,
                       expected_symbols=None):
        input_str = context.input_str
        position = context.position
        layout_content = context.layout_content
        h_print("Position:",
                pos_to_line_col(input_str, position), level=1)
        h_print("Context:", _(position_context(context)), level=1)
        if layout_content:
            h_print("Layout: ", "'{}'".format(_(layout_content)), level=1)
        if expected_symbols:
            h_print("Symbols expected: ",
                    [s.name for s in expected_symbols], level=1)
        if lookahead_tokens:
            h_print("Token(s) ahead:", _(str(lookahead_tokens)), level=1)

    @no_colors
    def _trace_head(self, new_head, label):
        self.dot_trace += '{} [label="{}"];\n'.format(new_head.key, label)

    @no_colors
    def _trace_step(self, old_head, new_head, root_head, label=''):
        new_head_key = new_head.key if isinstance(new_head, GSSNode) \
                       else new_head
        self.dot_trace += '{} -> {} [label="{}. {}" {}];\n'.format(
            old_head.key, new_head_key, self.debug_step, label,
            TRACE_DOT_STEP_STYLE)
        self.dot_trace += '{} -> {};\n'.format(new_head_key, root_head.key)

    @no_colors
    def _trace_step_finish(self, from_head):
        self._trace_step(from_head, "success", from_head)

    @no_colors
    def _trace_step_kill(self, from_head):
        self.dot_trace += \
            '{}_killed [shape="diamond" fillcolor="red" label="killed"];\n'\
            .format(from_head.key)
        self.dot_trace += '{} -> {}_killed [label="{}." {}];\n'\
            .format(from_head.key, from_head.key, self.debug_step,
                    TRACE_DOT_STEP_STYLE)

    @no_colors
    def _trace_step_drop(self, from_head, to_head):
        self.dot_trace += '{} -> {} [label="drop empty" {}];\n'\
            .format(from_head.key, to_head.key, TRACE_DOT_DROP_STYLE)

    def _export_dot_trace(self):
        file_name = "{}_trace.dot".format(self.file_name) \
                    if self.file_name else "parglare_trace.dot"
        with codecs.open(file_name, 'w', encoding="utf-8") as f:
            f.write(DOT_HEADER)
            f.write(self.dot_trace)
            f.write("}\n")

        prints("Generated file {}.".format(file_name))
        prints("You can use dot viewer or generate pdf with the "
               "following command:")
        h_print("dot -Tpdf {0} -O {0}.pdf".format(file_name))


class GSSNode(object):
    """Graphs Structured Stack node.

    Attributes:
        context(Context): The parsing context.
        any_empty(bool): If some of this node parent links results are empty.
        all_empty(bool): If all of this node parent link results are empty.
        parents(list): list of (parent GSSNode, result, any_empty, all_empty)
             Each stack node might have multiple parents which represent
             multiple path parse took to reach the current state. Each
             parent link keeps a result of semantic action executed during
             shift or reduce operation that created this node/link and the
             flag if any part of the result is obtained using epsilon/empty
             production.
    """
    __slots__ = ['context', 'parents', 'any_empty', 'all_empty',
                 'number_of_trees', '_hash']

    def __init__(self, context, number_of_trees=0):
        self.context = context

        # Initialize to neutral elements
        self.any_empty = False
        self.all_empty = True

        self.parents = []
        self.number_of_trees = number_of_trees

        self._hash = hash((self.context.state.state_id,
                           self.context.start_position,
                           self.context.token_ahead))

    def less_empty(self, other):
        return (other.all_empty and not self.all_empty) or \
            (other.any_empty and not self.any_empty)

    def merge_head(self, other, parser):
        """Merge same top stack nodes.

        Merge will be succesfull only if this node is "more empty" than the
        other node.
        """
        # Reject merging if other node is "more empty"
        if other.all_empty or self.less_empty(other):
            if parser.debug:
                h_print("Rejected merging of empty head: ", other, level=1)
            return False
        else:
            if self.parents and other.less_empty(self):
                if parser.debug:
                    h_print("Merging less empty head to more empty",
                            " -> less empty head wins.", level=1)
                    if parser.debug_trace:
                        for p in self.parents:
                            parser._trace_step_drop(self, p[0])
                self.any_empty = False
                self.all_empty = True
                self.parents = []
                self.number_of_trees = 0

            self.any_empty |= other.any_empty
            self.all_empty &= other.all_empty
            self.number_of_trees += other.number_of_trees
            self.parents.extend(other.parents)

            if parser.debug:
                h_print("Merging head ", other, level=1)
                h_print("to head", self, level=2)
            return True

    def create_link(self, parent, result, any_empty, all_empty, parser):
        self.parents.append((parent, result, any_empty, all_empty))
        self.number_of_trees += parent.number_of_trees
        self.any_empty |= any_empty
        self.all_empty &= all_empty
        if parser.debug:
            h_print("Creating link \tfrom head:", self, level=2)
            h_print("  to head:", parent, level=4)

    def for_token(self, token):
        """Create head for the given token either by returning this head if the
        token is appropriate or making a clone.

        This is used to support lexical ambiguity. Multiple tokens might be
        matched at the same state and position. In this case parser should
        fork and this is done by cloning stack head.

        """
        if self.context.token_ahead is None:
            self.context.token_ahead = token
            return self
        elif self.context.token_ahead == token:
            return self
        else:
            context = copy(self.context)
            context.token_ahead = token
            new_head = GSSNode(context, self.number_of_trees)
            new_head.parents = list(self.parents)
            new_head.any_empty = self.any_empty
            new_head.all_empty = self.all_empty
            return new_head

    def __eq__(self, other):
        """Stack nodes are equal if they are on the same position in the
        same state for the same lookahead token.

        """
        return self.context.state.state_id == other.context.state.state_id \
            and self.context.start_position == other.context.start_position \
            and self.context.token_ahead == other.context.token_ahead

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return _("<state={}:{}, id={}, pos={}, endpos={}{}, empty=[{},{}], "
                 "parents={}, trees={}>".format(
                     self.context.state.state_id, self.context.state.symbol,
                     id(self),
                     self.context.start_position, self.context.end_position,
                     ", token ahead={}".format(self.context.token_ahead)
                     if self.context.token_ahead is not None else "",
                     self.any_empty, self.all_empty, len(self.parents),
                     self.number_of_trees))

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return self._hash

    @property
    def key(self):
        """Head unique idenfier used for dot trace."""
        return "head_{}_{}_{}".format(self.context.state.state_id,
                                      self.context.start_position,
                                      self.context.end_position)


DOT_HEADER = """
    digraph parglare_trace {
    rankdir=LR
    fontname = "Bitstream Vera Sans"
    fontsize = 8
    node[
        style=filled,
        fillcolor=aliceblue
    ]
    nodesep = 0.3
    edge[dir=black,arrowtail=empty]

"""

TRACE_DOT_STEP_STYLE = 'color="red" style="dashed"'
TRACE_DOT_DROP_STYLE = 'color="orange" style="dotted"'



"""
Allow arbitrary user meta-data on both production and rule-level.
See issue: https://github.com/igordejanovic/parglare/issues/57
"""
import pytest
from parglare import Grammar, ParseError
from parglare.grammar import ASSOC_LEFT, ASSOC_RIGHT, ASSOC_NONE


def test_production_meta_data():

    grammar_str = r'''
    MyRule: 'a' {left, 1, dynamic, nops,
                 some_string:'My Label with \\ and \' end',
                 some_bool: true,
                 some_int: 3,
                 some_float: 4.5};
    '''

    grammar = Grammar.from_string(grammar_str)
    my_rule = grammar.get_nonterminal('MyRule')

    prod = my_rule.productions[0]
    assert prod.assoc == ASSOC_LEFT
    assert prod.prior == 1
    assert prod.dynamic
    assert prod.some_string == r"My Label with \ and ' end"
    assert prod.some_bool is True
    assert prod.some_int == 3
    assert prod.some_float == 4.5

    with pytest.raises(AttributeError):
        prod.non_existing


def test_production_meta_data_must_be_key_value():

    grammar_str = r'''
    MyRule: 'a' {left, 1, dynamic, nops, label:'My Label', not_allowed};
    '''

    with pytest.raises(ParseError, match=r'ot_allowed\*}'):
        Grammar.from_string(grammar_str)


def test_terminal_meta_data():

    grammar_str = r'''
    MyRule: a;
    terminals
    a: 'a' {dynamic, 1, label: 'My Label'};
    '''

    grammar = Grammar.from_string(grammar_str)
    term_a = grammar.get_terminal('a')

    assert term_a.prior == 1
    assert term_a.dynamic
    assert term_a.label == 'My Label'

    with pytest.raises(AttributeError):
        term_a.non_existing


def test_terminal_meta_data_must_be_key_value():

    grammar_str = r'''
    MyRule: a;
    terminals
    a: 'a' {dynamic, 1, label: 'My Label', not_allowed};
    '''

    with pytest.raises(ParseError, match=r'ot_allowed\*}'):
        Grammar.from_string(grammar_str)


def test_rule_meta():

    grammar_str = r'''
    MyRule {label: 'My Label', nops}: 'a' {left, 1, dynamic};
    '''

    grammar = Grammar.from_string(grammar_str)
    my_rule = grammar.get_nonterminal('MyRule')

    # User meta-data is accessible on non-terminal
    assert my_rule.label == 'My Label'

    # But built-in disambiguation rules are not
    with pytest.raises(AttributeError):
        assert my_rule.nops

    # Also for nonexisting attributes
    with pytest.raises(AttributeError):
        assert my_rule.nonexisting

    # Production has its own meta-data
    prod = my_rule.productions[0]
    assert prod.assoc == ASSOC_LEFT
    assert prod.prior == 1
    assert prod.dynamic

    # Rule-level meta-data are propagated to productions
    assert prod.label == 'My Label'


def test_rule_meta_override():
    """
    Test that meta-data are propagated to productions and can be overriden.
    """

    grammar_str = r'''
    MyRule {label: 'My Label', left}: 'a' {right, label: 'My overriden label'}
                                    | 'b';
    '''

    grammar = Grammar.from_string(grammar_str)
    my_rule = grammar.get_nonterminal('MyRule')

    # User meta-data is accessible on non-terminal
    assert my_rule.label == 'My Label'

    prod = my_rule.productions[0]
    # First production overrides meta-data
    assert prod.label == 'My overriden label'
    assert prod.assoc == ASSOC_RIGHT

    # If not overriden it uses meta-data from the rule.
    prod = my_rule.productions[1]
    assert prod.label == 'My Label'
    assert prod.assoc == ASSOC_LEFT


def test_multiple_rule_meta_override():
    """
    Test that meta-data are propagated to productions from containing rule
    and can be overriden.
    """

    grammar_str = r'''
    MyRule {label: 'My Label', left}: 'first' {right,
                                               label: 'My overriden label'}
                                    | 'second';

    MyRule {label: 'Other rule'}: 'third' {left}
                                | 'fourth' {label: 'Fourth prod'};
    '''

    grammar = Grammar.from_string(grammar_str)
    my_rule = grammar.get_nonterminal('MyRule')

    # User meta-data is accessible on non-terminal
    # Rule level meta-data are only those defined on the
    # first rule in the order of the definition.
    assert my_rule.label == 'My Label'

    assert len(my_rule.productions) == 4

    prod = my_rule.productions[0]
    # First production overrides meta-data
    assert prod.label == 'My overriden label'
    assert prod.assoc == ASSOC_RIGHT

    # If not overriden it uses meta-data from the rule.
    prod = my_rule.productions[1]
    assert prod.label == 'My Label'
    assert prod.assoc == ASSOC_LEFT

    # Third and fourth production belongs to the second rule so they
    # inherits its meta-data.
    prod = my_rule.productions[2]
    assert prod.label == 'Other rule'
    assert prod.assoc == ASSOC_LEFT

    prod = my_rule.productions[3]
    assert prod.label == 'Fourth prod'
    assert prod.assoc == ASSOC_NONE



import pytest  # noqa
import subprocess
import os

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
GRAMMAR_FILE = os.path.join(CURRENT_DIR, 'grammar.pg')


@pytest.mark.skipif("TRAVIS" in os.environ
                    and os.environ["TRAVIS"] == "true",
                    reason="Test fails under TRAVIS")
def test_pglr_check():
    """
    Test pglr command for grammar checking.
    """
    result = subprocess.call(['pglr', 'compile', GRAMMAR_FILE])
    assert result == 0


@pytest.mark.skipif("TRAVIS" in os.environ
                    and os.environ["TRAVIS"] == "true",
                    reason="Test fails under TRAVIS")
def test_pglr_viz():
    """
    Test pglr command for PDA visualization.
    """
    DOT_FILE = os.path.join(CURRENT_DIR, '{}.dot'.format(GRAMMAR_FILE))
    try:
        os.remove(DOT_FILE)
    except Exception:
        pass
    assert not os.path.exists(DOT_FILE)
    result = subprocess.call(['pglr', '--no-colors', 'viz', GRAMMAR_FILE])
    assert result == 0
    assert os.path.exists(DOT_FILE)
    assert 'digraph grammar' in open(DOT_FILE, 'r').read()



import os
from parglare import Grammar, Parser

this_folder = os.path.dirname(__file__)


def test_diamond_import_resolving_and_model_creation():
    g = Grammar.from_file(os.path.join(this_folder, 'model.pg'))
    assert g
    assert g.get_terminal('packages.components.base.COMMA')
    assert g.get_nonterminal('Model')

    # First path used for import of Component is going
    # packages->components->Component
    component_nonterminal = g.get_nonterminal('packages.components.Component')
    assert component_nonterminal

    input_str = '''

    package First
    package Second {
        component packageComponent {

        }
    }

    module SomeModule {

        component myComponent {
            in SomeInputSlot
            out SomeOutputSlot
        }


    }

    '''

    model = Parser(g).parse(input_str)
    assert model
    assert model.__class__.__name__ == 'Model'
    assert type(model.packages) is list
    assert len(model.packages) == 2
    assert model.packages[0].name == 'First'
    assert type(model.modules) is list
    assert len(model.modules) == 1

    packageComponent = model.packages[1].body.components[0]
    assert packageComponent.name == 'packageComponent'

    module = model.modules[0]
    assert module.__class__.__name__ == 'm.Module'
    assert module.name == 'SomeModule'
    assert len(module.components) == 1

    component = module.components[0]
    assert type(component) == type(packageComponent)
    assert component.name == 'myComponent'
    assert len(component.slots) == 2

    slot = component.slots[1]
    assert slot.__class__.__name__ == 'packages.components.SlotOut'
    assert slot.name == 'SomeOutputSlot'



import pytest  # noqa
import os
import tempfile
from parglare import Grammar
from parglare.tables import create_table
from parglare.export import grammar_pda_export


def test_dot_export():
    grammar = 'S: S S | S S S | "b";'
    g = Grammar.from_string(grammar)

    table = create_table(g)

    tmp_dir = tempfile.mkdtemp()
    file_name = os.path.join(tmp_dir, 'testexport.dot')

    grammar_pda_export(table, file_name)

    with open(file_name) as f:
        assert 'label' in f.read()

    os.remove(file_name)
    os.rmdir(tmp_dir)



from __future__ import unicode_literals
import re
from parglare import get_collector

recognizer = get_collector()
number_re = re.compile(r'\d+(\.\d+)*')
fqn_re = re.compile(r'\w+(\.\w+)*')


@recognizer('NUMERIC_ID')
def number(input, pos):
    number_match = number_re.match(input[pos:])
    if number_match:
        return input[pos:pos + len(number_match.group())]


@recognizer
def FQN(input, pos):
    fqn_match = fqn_re.match(input[pos:])
    if fqn_match:
        return input[pos:pos + len(fqn_match.group())]



from parglare import get_collector

recognizer = get_collector()


@recognizer
def A(input, pos):
    return [input[pos]]


# This should raise an exception as there is no `B` terminal in the grammar
@recognizer
def B(input, pos):
    return [input[pos]]



#!/usr/bin/env python
from __future__ import unicode_literals, print_function
import sys
import click
from parglare import Grammar, ParseError, GrammarError, GLRParser
from parglare.export import grammar_pda_export
from parglare.tables import create_load_table
from parglare.termui import prints, a_print, h_print
import parglare.termui as t


@click.group()
@click.option('--debug', default=False, is_flag=True,
              help="Debug/trace output.")
@click.option('--no-colors', default=False, is_flag=True,
              help="Disable output coloring.")
@click.option('--prefer-shifts', default=False, is_flag=True,
              help="Prefer shifts over reductions.")
@click.option('--prefer-shifts-over-empty',
              default=False, is_flag=True,
              help="Prefer shifts over empty reductions.")
@click.pass_context
def pglr(ctx, debug, no_colors, prefer_shifts, prefer_shifts_over_empty):
    """
    Command line interface for working with parglare grammars.
    """
    ctx.obj = {'debug': debug, 'colors': not no_colors,
               'prefer_shifts': prefer_shifts,
               'prefer_shifts_over_empty': prefer_shifts_over_empty}


@pglr.command()
@click.argument('grammar_file', type=click.Path())
@click.pass_context
def compile(ctx, grammar_file):
    debug = ctx.obj['debug']
    colors = ctx.obj['colors']
    prefer_shifts = ctx.obj['prefer_shifts']
    prefer_shifts_over_empty = ctx.obj['prefer_shifts_over_empty']
    h_print('Compiling...')
    compile_get_grammar_table(grammar_file, debug, colors, prefer_shifts,
                              prefer_shifts_over_empty)


@pglr.command()
@click.argument('grammar_file', type=click.Path())
@click.pass_context
def viz(ctx, grammar_file):
    debug = ctx.obj['debug']
    colors = ctx.obj['colors']
    prefer_shifts = ctx.obj['prefer_shifts']
    prefer_shifts_over_empty = ctx.obj['prefer_shifts_over_empty']
    t.colors = colors
    grammar, table = compile_get_grammar_table(grammar_file, debug, colors,
                                               prefer_shifts,
                                               prefer_shifts_over_empty)
    prints("Generating '%s.dot' file for the grammar PDA." % grammar_file)
    prints("Use dot viewer (e.g. xdot) "
           "or convert to pdf by running 'dot -Tpdf -O %s.dot'" % grammar_file)
    t.colors = False
    grammar_pda_export(table, "%s.dot" % grammar_file)


@pglr.command()
@click.argument('grammar_file', type=click.Path())
@click.option('--input-file', '-f', type=click.Path(),
              help="Input file for tracing")
@click.option('--input', '-i', help="Input string for tracing")
@click.pass_context
def trace(ctx, grammar_file, input_file, input):
    if not (input_file or input):
        prints('Expected either input_file or input string.')
        sys.exit(1)
    colors = ctx.obj['colors']
    prefer_shifts = ctx.obj['prefer_shifts']
    prefer_shifts_over_empty = ctx.obj['prefer_shifts_over_empty']
    grammar, table = compile_get_grammar_table(grammar_file, True, colors,
                                               prefer_shifts,
                                               prefer_shifts_over_empty)
    parser = GLRParser(grammar, debug=True, debug_trace=True,
                       debug_colors=colors, prefer_shifts=prefer_shifts,
                       prefer_shifts_over_empty=prefer_shifts_over_empty)
    if input:
        parser.parse(input)
    else:
        parser.parse_file(input_file)


def compile_get_grammar_table(grammar_file, debug, colors, prefer_shifts,
                              prefer_shifts_over_empty):
    try:
        g = Grammar.from_file(grammar_file, _no_check_recognizers=True,
                              debug_colors=colors)
        if debug:
            g.print_debug()
        table = create_load_table(
            g, prefer_shifts=prefer_shifts,
            prefer_shifts_over_empty=prefer_shifts_over_empty,
            force_create=True)
        if debug or table.sr_conflicts or table.rr_conflicts:
            table.print_debug()

        if not table.sr_conflicts and not table.rr_conflicts:
            h_print("Grammar OK.")

        if table.sr_conflicts:
            if len(table.sr_conflicts) == 1:
                message = 'There is 1 Shift/Reduce conflict.'
            else:
                message = 'There are {} Shift/Reduce conflicts.'\
                          .format(len(table.sr_conflicts))
            a_print(message)
            prints("Either use 'prefer_shifts' parser mode, try to resolve "
                   "manually, or use GLR parsing.".format(
                       len(table.sr_conflicts)))
        if table.rr_conflicts:
            if len(table.rr_conflicts) == 1:
                message = 'There is 1 Reduce/Reduce conflict.'
            else:
                message = 'There are {} Reduce/Reduce conflicts.'\
                          .format(len(table.rr_conflicts))
            a_print(message)
            prints("Try to resolve manually or use GLR parsing.")

    except (GrammarError, ParseError) as e:
        print("Error in the grammar file.")
        print(e)
        sys.exit(1)

    return g, table


if __name__ == '__main__':
    pglr()



# -*- coding: utf-8 -*-
# flake8: NOQA
from parglare.parser import Parser, Token, pos_to_line_col, \
    Node, NodeTerm, NodeNonTerm
from parglare.tables import LALR, SLR, SHIFT, REDUCE, ACCEPT
from parglare.glr import GLRParser
from parglare.grammar import Grammar, NonTerminal, Terminal, \
    RegExRecognizer, StringRecognizer, EMPTY, EOF, STOP
from parglare.common import get_collector
from parglare.exceptions import ParserInitError, ParseError, GrammarError, \
    DisambiguationError


__version__ = "0.9.0"



from parglare import get_collector

action = get_collector()


@action
def Calc(_, nodes):
    return nodes[-1]


@action
def Assignment(context, nodes):
    var_name, _, value = nodes
    context.extra[var_name] = float(value)


E = [
    lambda _, n: n[0] + n[2],
    lambda _, n: n[0] - n[2],
    lambda _, n: n[0] * n[2],
    lambda _, n: n[0] / n[2],
    lambda _, n: n[1],
    lambda context, n: context.extra[n[0]],
    lambda _, n: float(n[0])
]
action('E')(E)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pytest  # noqa
from parglare import Parser, GLRParser, Grammar, ParseError

parsers = pytest.mark.parametrize("parser_class", [Parser, GLRParser])


@parsers
def test_layout_whitespaces(parser_class):
    grammar = r"""
    S: K EOF;
    K: A | B;
    A: 'a' A | 'a';
    B: 'b' B | 'b';
    LAYOUT: WS | EMPTY;

    terminals
    WS: /\s+/;
    """
    g = Grammar.from_string(grammar)

    in_str = """aaa a    aaaa
    aa    aa a aaa

    aaa
    """

    parser = parser_class(g, debug=True)
    result = parser.parse(in_str)
    assert result


@parsers
def test_layout_simple_comments(parser_class):
    grammar = r"""
    S: K EOF;
    K: A | B;
    A: 'a' A | 'a';
    B: 'b' B | 'b';

    LAYOUT: LayoutItem | LAYOUT LayoutItem;
    LayoutItem: WS | Comment | EMPTY;

    terminals
    WS: /\s+/;
    Comment: /\/\/.*/;
    """
    g = Grammar.from_string(grammar)

    in_str = """aaa a    aaaa
    aa    aa a aaa // This is a comment

    aaa
    """

    parser = parser_class(g)
    parser.parse(in_str)


@parsers
def test_layout_nested_comments(parser_class):
    grammar = r"""
    S: K EOF;
    K: 'a' B | 'a' C;
    B: 'b' | B 'b';
    C: 'c' | C 'c';

    LAYOUT: LayoutItem | LAYOUT LayoutItem;
    LayoutItem: WS | Comment | EMPTY;
    Comment: '/*' CorNCs '*/' | LineComment;
    CorNCs: CorNC | CorNCs CorNC | EMPTY;
    CorNC: Comment | NotComment | WS;

    terminals
    WS: /\s+/;
    LineComment: /\/\/.*/;
    NotComment: /((\*[^\/])|[^\s*\/]|\/[^\*])+/;
    """
    g = Grammar.from_string(grammar)

    in_str = """//Line comment at beginning
    a  b b b   b // This is line comment
    b b b b b b  /* This is block
    comment */

    bbbb  b b b b b
    /* Another block comment
       // With nested line comment
       /* And nested block
    comment */
    */

    bbbb b b b
    """

    parser = parser_class(g)
    parser.parse(in_str)


@parsers
def test_layout_context(parser_class):
    """
    Test that layout is passed in the action context.
    """
    grammar = r"""
    S: K EOF;
    K: A | B;
    A: 'a' A | 'a';
    B: 'b' B | 'b';

    LAYOUT: LayoutItem | LAYOUT LayoutItem;
    LayoutItem: WS | Comment | EMPTY;

    terminals
    WS: /\s+/;
    Comment: /\/\/.*/;
    """
    g = Grammar.from_string(grammar)

    in_str = """aaa a    aaaa
    aa    aa a aaa // This is a comment

    aaa
    """

    def layout_action(context, _):
        layout_called[0] = True
        if 'This is a comment' in context.layout_content:
            layout_passed[0] = True

    actions = {
        "a": layout_action
    }

    layout_called = [False]
    layout_passed = [False]

    parser = parser_class(g, actions=actions)

    parser.parse(in_str)

    assert layout_called[0]
    assert layout_passed[0]


@parsers
def test_layout_actions(parser_class):
    """
    Test that user provided actions for layout rules are called if given.
    """

    grammar = r"""
    S: K EOF;
    K: A | B;
    A: 'a' A | 'a';
    B: 'b' B | 'b';

    LAYOUT: LayoutItem | LAYOUT LayoutItem;
    LayoutItem: WS | Comment | EMPTY;

    terminals
    WS: /\s+/;
    Comment: /\/\/.*/;
    """
    g = Grammar.from_string(grammar)

    in_str = """aaa a    aaaa
    aa    aa a aaa // This is a comment

    aaa
    """

    def comment_action(_, nodes):
        layout_called[0] = True

    def layout_action(_, nodes):
        ret = ''
        for n in nodes:
            if n:
                ret += n
        return ret

    def a_action(context, _):
        called[0] = True

    actions = {
        'Comment': comment_action,
        'LAYOUT': layout_action,
        'a': a_action
    }

    called = [False]
    layout_called = [False]

    parser = parser_class(g, actions=actions, layout_actions=actions)
    parser.parse(in_str)

    assert called[0]
    assert layout_called[0]


def test_layout_terminal():
    """
    Test that layout definition may be just a terminal rule.
    """
    grammar = r"""
    S: "a" "b";
    LAYOUT: "c";
    """

    g = Grammar.from_string(grammar)
    parser = Parser(g)

    with pytest.raises(ParseError):
        parser.parse("a b")
    parser.parse("cacbc")

    grammar = r"""
    S: "a" "b";
    LAYOUT: DIGITS;

    terminals
    DIGITS: /\d*/;
    """

    g = Grammar.from_string(grammar)
    parser = Parser(g)

    with pytest.raises(ParseError):
        parser.parse("a b")
    result = parser.parse("4444a23b545")
    assert result == ['a', 'b']



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pytest
from parglare import GLRParser, Grammar, Parser, ParseError
from parglare.exceptions import SRConflicts


def test_lr2_grammar():

    grammar = r"""
    Model: Prods EOF;
    Prods: Prod | Prods Prod;
    Prod: ID "=" ProdRefs;
    ProdRefs: ID | ProdRefs ID;

    terminals
    ID: /\w+/;
    """

    input_str = """
    First = One Two three
    Second = Foo Bar
    Third = Baz
    """

    g = Grammar.from_string(grammar)

    # This grammar is not LR(1) as it requires
    # at least two tokens of lookahead to decide
    # what to do on each ID from the right side.
    # If '=' is after ID than it should reduce "Prod"
    # else it should reduce ID as ProdRefs.
    with pytest.raises(SRConflicts):
        Parser(g, prefer_shifts=False)

    # prefer_shifts strategy (the default)
    # will remove conflicts but the resulting parser
    # will fail to parse any input as it will consume
    # greadily next rule ID as the body element of the previous Prod rule.
    parser = Parser(g)
    with pytest.raises(ParseError):
        parser.parse(input_str)

    # But it can be parsed unambiguously by GLR.
    p = GLRParser(g)

    results = p.parse(input_str)
    assert len(results) == 1


def test_nops():
    """
    Test that nops (no prefer shifts) will honored per rule.
    """
    grammar = """
    Program: "begin"
             statements=Statements
             ProgramEnd EOF;
    Statements: Statements1 | EMPTY;
    Statements1: Statements1 Statement | Statement;
    ProgramEnd: End;
    Statement: End "transaction" | "command";

    terminals
    End: "end";
    """

    g = Grammar.from_string(grammar, ignore_case=True)
    parser = GLRParser(g, build_tree=True, prefer_shifts=True)

    # Here we have "end transaction" which is a statement and "end" which
    # finish program. Prefer shift strategy will make parser always choose to
    # shift "end" in anticipation of "end transaction" statement instead of
    # reducing by "Statements" and finishing.
    with pytest.raises(ParseError):
        parser.parse("""
        begin
            command
            end transaction
            command
            end transaction
            command
        end
        """)

    # When {nops} is used, GLR parser will investigate both possibilities at
    # this place and find the correct interpretation while still using
    # prefer_shift strategy globaly.
    grammar = """
    Program: "begin"
             statements=Statements
             ProgramEnd EOF;
    Statements: Statements1 {nops} | EMPTY;
    Statements1: Statements1 Statement | Statement;
    ProgramEnd: End;
    Statement: End "transaction" | "command";

    terminals
    End: "end";
    """

    g = Grammar.from_string(grammar, ignore_case=True)
    parser = GLRParser(g, build_tree=True, prefer_shifts=True)
    parser.parse("""
    begin
        command
        end transaction
        command
        end transaction
        command
    end
    """)


def test_expressions():

    actions = {
        "s": lambda _, c: c[0],
        "E": [
            lambda _, nodes: nodes[0] + nodes[2],
            lambda _, nodes: nodes[0] * nodes[2],
            lambda _, nodes: nodes[1],
            lambda _, nodes: int(nodes[0])
        ]
    }

    # This grammar is highly ambiguous if priorities and
    # associativities are not defined to disambiguate.
    grammar = r"""
    s: E EOF;
    E: E "+" E | E "*" E | "(" E ")" | Number;
    terminals
    Number: /\d+/;
    """
    g = Grammar.from_string(grammar)
    p = GLRParser(g, actions=actions, debug=True)

    # Even this simple expression has 2 different interpretations
    # (4 + 2) * 3 and
    # 4 + (2 * 3)
    results = p.parse("4 + 2 * 3")
    assert len(results) == 2
    assert 18 in results and 10 in results

    # Adding one more operand rises number of interpretations to 5
    results = p.parse("4 + 2 * 3 + 8")
    assert len(results) == 5

    # One more and there are 14 interpretations
    results = p.parse("4 + 2 * 3 + 8 * 5")
    assert len(results) == 14

    # The number of interpretation will be the Catalan number of n
    # where n is the number of operations.
    # https://en.wikipedia.org/wiki/Catalan_number
    # This number rises very fast. For 10 operations number of interpretations
    # will be 16796!

    # If we rise priority for multiplication operation we reduce ambiguity.
    # Default production priority is 10. Here we will raise it to 15 for
    # multiplication.
    grammar = r"""
    s: E EOF;
    E: E "+" E | E "*" E {15}| "(" E ")" | Number;
    terminals
    Number: /\d+/;
    """
    g = Grammar.from_string(grammar)
    p = GLRParser(g, actions=actions)

    # This expression now has 2 interpretation:
    # (4 + (2*3)) + 8
    # 4 + ((2*3) + 8)
    # This is due to associativity of + operation which is not defined.
    results = p.parse("4 + 2 * 3 + 8")
    assert len(results) == 2

    # If we define associativity for both + and * we have resolved all
    # ambiguities in the grammar.
    grammar = r"""
    s: E EOF;
    E: E "+" E {left}| E "*" E {left, 15}| "(" E ")" | Number;
    terminals
    Number: /\d+/;
    """
    g = Grammar.from_string(grammar)
    p = GLRParser(g, actions=actions)

    results = p.parse("4 + 2 * 3 + 8 * 5 * 3")
    assert len(results) == 1
    assert results[0] == 4 + 2 * 3 + 8 * 5 * 3


def test_epsilon_grammar():

    grammar = r"""
    Model: Prods EOF;
    Prods: Prod | Prods Prod | EMPTY;
    Prod: ID "=" ProdRefs;
    ProdRefs: ID | ProdRefs ID;

    terminals
    ID: /\w+/;
    """

    g = Grammar.from_string(grammar)
    p = GLRParser(g, debug=True)

    txt = """
    First = One Two three
    Second = Foo Bar
    Third = Baz
    """

    results = p.parse(txt)
    assert len(results) == 1

    results = p.parse("")
    assert len(results) == 1


def test_non_eof_grammar_nonempty():
    """
    Grammar that is not anchored by EOF at the end might
    result in multiple trees that are produced by sucessful
    parses of the incomplete input.
    """
    grammar_nonempty = r"""
    Model: Prods;
    Prods: Prod | Prods Prod;
    Prod: ID "=" ProdRefs;
    ProdRefs: ID | ProdRefs ID;

    terminals
    ID: /\w+/;
    """

    g_nonempty = Grammar.from_string(grammar_nonempty)

    txt = """
    First = One Two three
    Second = Foo Bar
    Third = Baz
    """

    p = GLRParser(g_nonempty, debug=True)
    results = p.parse(txt)
    # There are eight succesful parses:
    # 1. First = One
    # 2. First = One Two
    # 3. First = One Two three
    # 4. First = One Two three Second
    # 5. ... Second = Foo
    # 6. ... Second = Foo Bar
    # 7. ... Second = Foo Bar Third
    # 8. everyting parsed
    assert len(results) == 8

    # With lexical disambiguation turned on there are only 3 parses,
    # because regular tokens are preferred over STOP stop tokens.
    # So STOP token gets fed to the parser head only if there is nothing else.
    # This is the situation when head parsing ProdRefs encounters "=".
    # Namely the three parses are:
    # 1. First = One Two three Second
    # 2. ... Second = Foo Bar Third
    # 3. everything parsed
    disambig_p = GLRParser(g_nonempty, lexical_disambiguation=True)
    assert len(disambig_p.parse(txt)) == 3


def test_non_eof_grammar_empty():
    """
    Grammar that is not anchored by EOF at the end might
    result in multiple trees that are produced by sucessful
    parses of the incomplete input.
    """
    grammar_empty = r"""
    Model: Prods;
    Prods: Prod | Prods Prod | EMPTY;
    Prod: ID "=" ProdRefs;
    ProdRefs: ID | ProdRefs ID;

    terminals
    ID: /\w+/;
    """

    g_empty = Grammar.from_string(grammar_empty)

    txt = """
    First = One Two three
    Second = Foo Bar
    Third = Baz
    """

    p = GLRParser(g_empty, debug=True)

    results = p.parse(txt)
    assert len(results) == 8

    results = p.parse("")
    assert len(results) == 1


def test_empty_terminal():
    g = Grammar.from_string("""
    start: a EOF;
    a: a t | t;
    terminals
    t: /b*/;
    """)
    p = GLRParser(g)
    with pytest.raises(ParseError):
        p.parse("a")


def test_empty_recognizer():
    """This test verifies if custom recorgnizer matching on empty string
    can throw parser into infinite loop."""

    def match_bs(input_str, pos):
        end_pos = pos
        while end_pos < len(input_str) and input_str[end_pos] == 'b':
            end_pos += 1
        return input_str[pos:end_pos]

    g = Grammar.from_string("""
    start: a EOF;
    a: a t | t;
    terminals
    t: ;
    """, recognizers={'t': match_bs})
    p = GLRParser(g)
    with pytest.raises(ParseError):
        p.parse("a")


def test_terminal_collision():
    g = Grammar.from_string("""
    expression: "1" s letter EOF
              | "2" s "A" EOF
              ;

    s: " ";

    terminals
    letter: /[A-Z]/;
    """)

    p = GLRParser(g, ws='')

    p.parse("2 A")
    p.parse("1 B")
    p.parse("1 A")


def test_lexical_ambiguity():
    g = Grammar.from_string("""
    expression: a "x" EOF
              | b EOF
              ;

    a: "x";
    b: "xx";
    """)

    p = GLRParser(g)

    assert sorted(p.parse("xx")) == [
        ['x', 'x', None],
        ['xx', None],
    ]

    disambig_p = GLRParser(g, lexical_disambiguation=True)

    assert sorted(disambig_p.parse("xx")) == [
        ['xx', None],
    ]



from __future__ import unicode_literals
from parglare import get_collector

action = get_collector()


@action('base.numeric')
def number(_, value):
    "This action is overriding by action name in 'base' module."
    return 43



"""
Common parsing actions.
"""
import sys

if sys.version < '3':
    text = unicode  # NOQA
else:
    text = str


def pass_none(_, value):
    return None


def pass_nochange(_, value):
    return value


def pass_empty(_, value):
    """
    Used for EMPTY production alternative in collect.
    """
    return []


def pass_single(_, nodes):
    """
    Unpack single value and pass up.
    """
    return nodes[0]


def pass_inner(_, nodes):
    """
    Pass inner value up, e.g. for parentheses '(' token ')'.
    """
    return nodes[1]


def collect_first(_, nodes):
    """
    Used for:
    Elements = Elements Element;
    """
    e1, e2 = nodes
    if e2 is not None:
        e1 = list(e1)
        e1.append(e2)
    return e1


def collect_first_sep(_, nodes):
    """
    Used for:
    Elements = Elements "," Element;
    """
    e1, _, e2 = nodes
    if e2 is not None:
        e1 = list(e1)
        e1.append(e2)
    return e1


def collect_right_first(_, nodes):
    """
    Used for:
    Elements = Element Elements;
    """
    e1, e2 = [nodes[0]], nodes[1]
    e1.extend(e2)
    return e1


def collect_right_first_sep(_, nodes):
    """
    Used for:
    Elements = Element "," Elements;
    """
    e1, e2 = [nodes[0]], nodes[2]
    e1.extend(e2)
    return e1


# Used for productions of the form - one or more elements:
# Elements: Elements Element | Element;
collect = [
    collect_first,
    pass_nochange
]

# Used for productions of the form - one or more elements:
# Elements: Elements "," Element | Element;
collect_sep = [
    collect_first_sep,
    pass_nochange
]

# Used for productions of the form - zero or more elements:
# Elements: Elements Element | Element | EMPTY;
collect_optional = [
    collect_first,
    pass_nochange,
    pass_empty
]

# Used for productions of the form - zero or more elements:
# Elements: Elements "," Element | Element | EMPTY;
collect_sep_optional = [
    collect_first_sep,
    pass_nochange,
    pass_empty
]

# Used for productions of the form - one or more elements:
# Elements: Element Elements | Element;
collect_right = [
    collect_right_first,
    pass_nochange
]

# Used for productions of the form - one or more elements:
# Elements: Element "," Elements | Element;
collect_right_sep = [
    collect_right_first_sep,
    pass_nochange
]

# Used for productions of the form - zero or more elements:
# Elements: Element Elements | Element | EMPTY;
collect_right_optional = [
    collect_right_first,
    pass_nochange,
    pass_empty
]

# Used for productions of the form - zero or more elements:
# Elements: Element "," Elements | Element | EMPTY;
collect_right_sep_optional = [
    collect_right_first_sep,
    pass_nochange,
    pass_empty
]

# Used for the production of the form:
# OptionalElement: Element | EMPTY;
optional = [
    pass_single,
    pass_none
]


def obj(context, nodes, **attrs):
    """
    Creates Python object with the attributes created from named matches.
    This action is used as default action for rules with named matches.
    """
    grammar = context.parser.grammar
    rule_name = context.production.symbol.fqn

    cls = grammar.classes[rule_name]
    instance = cls(**attrs)

    instance._pg_start_position = context.start_position
    instance._pg_end_position = context.end_position

    return instance



from parglare import Grammar, GLRParser


def test_issue31_glr_drop_parses_on_lexical_ambiguity():
    grammar = """
    model: element+ EOF;
    element: title
           | table_with_note
           | table_with_title;
    table_with_title: table_title table_with_note;
    table_with_note: table note*;

    terminals
    title: /title/;   // <-- This is lexically ambiguous with the next.
    table_title: /title/;
    table: "table";
    note: "note";
    """

    # this input should yield 4 parse trees.
    input = "title table title table"

    g = Grammar.from_string(grammar)
    parser = GLRParser(g, debug=True, debug_colors=True)
    results = parser.parse(input)

    # We should have 4 solutions for the input.
    assert len(results) == 4



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
import sys
from parglare.termui import s_attention as _a

if sys.version < '3':
    text = unicode  # NOQA
else:
    text = str


class Location(object):
    """
    Represents a location (point or span) of the object in the source code.

    Args:
    context(Context): Parsing context used to populate this object.
    line, column (int):
    """

    __slots__ = ['input_str', 'start_position', 'end_position', 'position',
                 'file_name', '_line', '_column']

    def __init__(self, context=None, file_name=None):

        self.input_str = context.input_str if context else None
        self.start_position = context.start_position if context else None
        self.end_position = context.end_position if context else None
        self.position = context.position if context else None
        if file_name:
            self.file_name = file_name
        elif context:
            self.file_name = context.file_name

        # Evaluate this only when string representation is needed.
        # E.g. during error reporting
        self._line = None
        self._column = None

    @property
    def line(self):
        if self._line is None:
            self.evaluate_line_col()
        return self._line

    @property
    def column(self):
        if self._column is None:
            self.evaluate_line_col()
        return self._column

    def evaluate_line_col(self):
        from parglare.parser import pos_to_line_col
        self._line, self._column = \
            pos_to_line_col(self.input_str, self.start_position)

    def __str__(self):
        line, column = self.line, self.column
        if line is not None:
            return _a('{}{}:{}:"{}"'
                      .format("{}:".format(self.file_name)
                              if self.file_name else "",
                              line, column,
                              position_context(self)))
        elif self.file_name:
            return _a(self.file_name)
        else:
            return "<Unknown location>"


def position_context(context):
    """
    Returns position context string.
    """
    start = max(context.position-10, 0)
    c = text(context.input_str[start:context.position]) + "*" \
        + text(context.input_str[context.position:context.position+10])
    return replace_newlines(c)


def replace_newlines(in_str):
    try:
        return in_str.replace("\n", "\\n")
    except AttributeError:
        return in_str


def load_python_module(mod_name, mod_path):
    """
    Loads Python module from an arbitrary location.
    See https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path  # noqa
    """
    if sys.version_info >= (3, 5):
        import importlib.util
        spec = importlib.util.spec_from_file_location(
            mod_name, mod_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    elif sys.version_info >= (3, 3):
        from importlib.machinery import SourceFileLoader
        module = SourceFileLoader(
            mod_name, mod_path).load_module()
    else:
        import imp
        module = imp.load_source(mod_name, mod_path)

    return module


def get_collector():
    """
    Produces action/recognizers collector/decorator that will collect all
    decorated objects under dictionary attribute `all`.
    """
    all = {}

    class Collector(object):
        def __call__(self, name_or_f):
            """
            If called with action/recognizer name return decorator.
            If called over function apply decorator.
            """
            is_name = type(name_or_f) in [str, text]

            def decorator(f):
                if is_name:
                    name = name_or_f
                else:
                    name = f.__name__
                objects = all.get(name, None)
                if objects:
                    if type(objects) is list:
                        objects.append(f)
                    else:
                        all[name] = [objects, f]
                else:
                    all[name] = f
                return f
            if is_name:
                return decorator
            else:
                return decorator(name_or_f)

    objects = Collector()
    objects.all = all
    return objects



from parglare import Grammar, NonTerminal, Terminal, RegExRecognizer


def get_grammar():

    # Expression grammar with float numbers
    E, T, F = [NonTerminal(name) for name in ['E', 'T', 'F']]
    PLUS, MULT, OPEN, CLOSE = [
        Terminal(value) for value in ['+', '*', '(', ')']]
    NUMBER = Terminal('number', RegExRecognizer(r'\d+(\.\d+)?'))
    productions = [
        (E, (E, PLUS, T)),
        (E, (T, )),
        (T, (T, MULT, F)),
        (T, (F, )),
        (F, (OPEN, E, CLOSE)),
        (F, (NUMBER,))
    ]

    return Grammar.from_struct(productions=productions, start_symbol=E)



from __future__ import unicode_literals
from parglare.termui import s_header as _


class LocationError(Exception):
    def __init__(self, location, message):
        self.location = location
        super(LocationError, self).__init__(
            "Error at {} => {}".format(location, message))


class GrammarError(LocationError):
    def __init__(self, location, message):
        location.position = location.start_position
        super(GrammarError, self).__init__(location, message)


class ParseError(LocationError):
    def __init__(self, location, symbols_expected, tokens_ahead=None,
                 symbols_before=None):
        self.symbols_expected = symbols_expected
        self.tokens_ahead = tokens_ahead if tokens_ahead else []
        self.symbols_before = symbols_before if symbols_before else []
        message = expected_message(symbols_expected, tokens_ahead)
        super(ParseError, self).__init__(location, message)


def expected_message(symbols_expected, tokens_ahead=None):
    return _('Expected: ') \
        + _(' or ').join(sorted([s.name for s in symbols_expected])) \
        + ((_(' but found ') +
            _(' or ').join(sorted([str(t) for t in tokens_ahead])))
           if tokens_ahead else '')


def expected_symbols_str(symbols):
    return " or ".join(sorted([s.name for s in symbols]))


def disambiguation_error(tokens):
    return 'Can\'t disambiguate between: {}'.format(
                _(' or ').join(sorted([str(t) for t in tokens])))


class ParserInitError(Exception):
    pass


class DisambiguationError(LocationError):
    def __init__(self, location, tokens):
        self.tokens = tokens
        message = disambiguation_error(tokens)
        super(DisambiguationError, self).__init__(location, message)


class DynamicDisambiguationConflict(Exception):
    def __init__(self, context, actions):
        self.state = state = context.state
        self.token = token = context.token
        self.actions = actions

        from parglare.parser import SHIFT
        message = "{}\nIn state {}:{} and input symbol '{}' after calling"\
                  " dynamic disambiguation still can't decide "\
                  .format(str(state), state.state_id, state.symbol, token)
        if actions[0].action == SHIFT:
            prod_str = " or ".join(["'{}'".format(str(a.prod))
                                    for a in actions[1:]])
            message += "whether to shift or reduce by "\
                       "production(s) {}.".format(prod_str)
        else:
            prod_str = " or ".join(["'{}'".format(str(a.prod))
                                    for a in actions])
            message += "which reduction to perform: {}".format(prod_str)

        self.message = message

    def __str__(self):
        return self.message


class LRConflict(object):
    def __init__(self, state, term, productions):
        self.message = ""
        self.state = state
        self.term = term
        self.productions = productions

    @property
    def dynamic(self):
        return self.term in self.state.dynamic


class SRConflict(LRConflict):
    def __init__(self, state, term, productions):
        super(SRConflict, self).__init__(state, term, productions)
        prod_str = " or ".join(["'{}'".format(str(p))
                                for p in productions])
        message = "{}\nIn state {}:{} and input symbol '{}' can't " \
                  "decide whether to shift or reduce by production(s) {}." \
            .format(str(state), state.state_id, state.symbol, term, prod_str)
        if self.dynamic:
            message += " Dynamic disambiguation strategy will be called."
        self.message = message


class RRConflict(LRConflict):
    def __init__(self, state, term, productions):
        super(RRConflict, self).__init__(state, term, productions)
        prod_str = " or ".join(["'{}'".format(str(p))
                                for p in productions])
        message = "{}\nIn state {}:{} and input symbol '{}' can't " \
                  "decide which reduction to perform: {}." \
                  .format(str(state), state.state_id, state.symbol, term,
                          prod_str)
        if self.dynamic:
            message += " Dynamic disambiguation strategy will be called."
        self.message = message


class LRConflicts(Exception):
    def __init__(self, conflicts):
        self.conflicts = conflicts
        message = "\n{} conflicts in following states: {}"\
                  .format(self.kind,
                          set([c.state.state_id for c in conflicts]))
        super(LRConflicts, self).__init__(message)


class SRConflicts(LRConflicts):
    kind = "Shift/Reduce"


class RRConflicts(LRConflicts):
    kind = "Reduce/Reduce"



import pytest
from parglare import Grammar, GrammarError


def test_conflict_between_string_match_and_rule_with_same_name():

    with pytest.raises(GrammarError, match=r'.*already defined.*'):
        Grammar.from_string('''
            Terminals:
                "Terminals" ":" terminal_list=Terminal* ";"
            ;
        ''')
